import os
import sys

sys.path.insert(0, os.path.abspath('../src'))

# meta
project = 'importloc'
copyright = '2025, Michael Makukha'
author = 'Michael Makukha'
release = '0.3.1'

# general
default_role = 'py:obj'
extensions = [
    'myst_parser',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx_design',
    'sphinx_sitemap',
]
nitpicky = True
suppress_warnings = [
    'myst.header',
    'myst.xref_missing',  # when generating TOC in usage.md
]
templates_path = ['_templates']

# autodoc
add_module_names = False
autoclass_content = 'both'
autodoc_docstring_signature = True
autodoc_member_order = 'bysource'
autodoc_typehints = 'signature'

# extlinks
extlinks = {
    'pypi': ('https://pypi.org/project/%s/', '%s'),
}

# intersphinx
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master', None),
}

# myst
myst_heading_anchors = 4

# html

html_baseurl = f'https://{project}.readthedocs.io/en/latest/'
html_theme = 'furo'
html_title = f'{project} {release}'
language = 'en'

html_css_files = ['custom.css']
html_static_path = ['_static']

html_theme_options = {
    'footer_icons': [
        {
            'name': 'GitHub',
            'url': f'https://github.com/makukha/{project}',
            'html': """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            'class': '',
        },
    ],
}



from .exc import InvalidLocation, ModuleNameConflict
from .location import ConflictResolution, Location, ModuleLocation, PathLocation
from .util import (
    get_instances,
    get_subclasses,
    getattr_nested,
    random_name,
)


__version__ = '0.3.1'

__all__ = [
    '__version__',
    'ConflictResolution',
    'InvalidLocation',
    'Location',
    'ModuleNameConflict',
    'ModuleLocation',
    'PathLocation',
    'get_instances',
    'get_subclasses',
    'getattr_nested',
    'random_name',
]



from typing import Any


class InvalidLocation(ValueError):
    """
    Incorrect format of location specification string.
    """


class ModuleNameConflict(ImportError):
    """
    Module with this name is already imported.
    """

    def __init__(self, modname: str, *args: Any, **kwargs: Any) -> None:
        msg = f'Module "{modname}" is already imported'
        super().__init__(msg, *args, **kwargs)



from dataclasses import dataclass
import doctest
import os
from pathlib import Path
import sys
from tempfile import TemporaryDirectory
from typing import Optional, Tuple, Union
import unittest

from importloc import Location


@dataclass
class File:
    path: str
    text: str


@dataclass
class DirectoryLayout:
    files: Tuple[File, ...]
    chdir: Union[str, Path] = '.'

    def __post_init__(self) -> None:
        self._tempdir: Optional[TemporaryDirectory[str]] = None
        self._oldcwd: Optional[str] = None

    def create(self) -> None:
        self._tempdir = TemporaryDirectory()
        for f in self.files:
            p = Path(self._tempdir.name) / f.path
            p.parent.mkdir(parents=True, exist_ok=True)
            p.write_text(f.text)

    def pushd(self) -> None:
        self._oldcwd = os.getcwd()
        os.chdir(self.cwd)

    @property
    def cwd(self) -> Path:
        if self._tempdir is None:
            raise RuntimeError('Directory layout must be created first')
        return Path(self._tempdir.name) / self.chdir

    def popd(self) -> None:
        if self._oldcwd is None:
            raise RuntimeError('Directory stack is already empty')
        os.chdir(self._oldcwd)
        self._oldcwd = None

    def destroy(self) -> None:
        if self._tempdir is None:
            raise RuntimeError('Directory layout already destroyed')
        self._tempdir.cleanup()
        self._tempdir = None


class TestCase(unittest.TestCase):
    layout: DirectoryLayout
    modnames: Tuple[str, ...]
    __test__ = False

    def __init_subclass__(cls, layout: DirectoryLayout) -> None:
        cls.layout = layout
        cls.__test__ = True

    def setUp(self) -> None:
        self.layout.create()
        self.layout.pushd()
        self.modnames = tuple(sys.modules.keys())
        sys.path.insert(0, str(self.layout.cwd))

    def tearDown(self) -> None:
        sys.path.pop(0)
        self.layout.popd()
        self.layout.destroy()
        for m in tuple(sys.modules):
            if m not in self.modnames:
                del sys.modules[m]
        self.modnames = ()

    def test_case(self) -> None:
        finder = doctest.DocTestFinder(recurse=False)
        runner = doctest.DocTestRunner(optionflags=doctest.ELLIPSIS | doctest.FAIL_FAST)
        globs = globals() | {'Location': Location}
        for test in finder.find(self.__doc__, self.__class__.__name__, globs=globs):
            ret = runner.run(test)
            self.assertFalse(ret.failed)



from dataclasses import replace

from importloc.testing import TestCase, DirectoryLayout, File


app_layout = DirectoryLayout(
    files=(
        File('app/__main__.py', 'def cli(): ...'),
        File('app/config.py', 'class Config:\n  class Nested: ...\nconf = Config()'),
        File(
            'app/errors.py',
            'class Error1(Exception): ...\nclass Error2(Exception): ...',
        ),
    ),
)


# various locations


class L1(TestCase, layout=app_layout):
    """Import from file

    ```python
    Location('app/config.py:conf').load()
    ```

    >>> loc = Location('app/config.py:conf')
    >>> loc
    <PathLocation 'app/config.py' obj='conf'>
    >>> loc.load()
    <config.Config object at 0x...>
    """


class L2(TestCase, layout=app_layout):
    """Import from module

    ```python
    Location('app.__main__:cli').load()
    ```

    >>> loc = Location('app.__main__:cli')
    >>> loc
    <ModuleLocation 'app.__main__' obj='cli'>
    >>> loc.load()
    <function cli at 0x...>
    """


class L3(TestCase, layout=replace(app_layout, chdir='app')):
    """Distinguish file and module locations

    ```python
    Location('./config.py:conf').load()
    ```

    >>> loc = Location('config.py:conf')
    >>> loc
    <ModuleLocation 'config.py' obj='conf'>
    >>> loc.load()
    Traceback (most recent call last):
        ...
    ModuleNotFoundError: No module named 'config.py'...

    Use relative path (similar to Docker bind mount). Path separator will result in
    `PathLocation` instead of `ModuleLocation`.

    >>> loc = Location('./config.py:conf')
    >>> loc
    <PathLocation 'config.py' obj='conf'>
    >>> loc.load()
    <config.Config object at 0x...>
    """


# various targets


class T1(TestCase, layout=app_layout):
    """Import nested class

    ```python
    Location('app/config.py:Config.Nested').load()
    ```

    >>> loc = Location('app/config.py:Config.Nested')
    >>> loc
    <PathLocation 'app/config.py' obj='Config.Nested'>
    >>> loc.load()
    <class 'config.Config.Nested'>
    """


class T2(TestCase, layout=app_layout):
    """Import module as a whole

    ```python
    Location('app/config.py').load()
    ```

    >>> loc = Location('app/config.py')
    >>> loc
    <PathLocation 'app/config.py'>
    >>> loc.load()
    <module 'config' from '...'>
    """


class T3(TestCase, layout=replace(app_layout, chdir='app')):
    """Use `Path` object when loading module

    ```python
    Location(Path('config.py')).load()
    ```

    >>> from pathlib import Path
    >>> loc = Location(Path('config.py'))
    >>> loc
    <PathLocation 'config.py'>
    >>> loc.load()
    <module 'config' from '...'>
    """


class T4(TestCase, layout=app_layout):
    """Import all instances of some type

    ```python
    get_instances(Location('app.__main__').load(), Callable)
    ```

    >>> from collections.abc import Callable
    >>> from importloc import get_instances
    >>> loc = Location('app.__main__')
    >>> loc
    <ModuleLocation 'app.__main__'>
    >>> get_instances(loc.load(), Callable)
    [<function cli at 0x...>]
    """


class T5(TestCase, layout=app_layout):
    """Import all subclasses

    ```python
    get_subclasses(Location('app.errors').load(), Exception)
    ```

    >>> from importloc import get_subclasses
    >>> loc = Location('app.errors')
    >>> loc
    <ModuleLocation 'app.errors'>
    >>> get_subclasses(loc.load(), Exception)
    [<class 'app.errors.Error1'>, <class 'app.errors.Error2'>]
    """


# override default module name


class N1(TestCase, layout=app_layout):
    """Use different module name

    ```python
    Location('...').load(modname='app_main')
    ```

    >>> Location('app/config.py:Config').load(modname='app_main')
    <class 'app_main.Config'>
    """


class N2(TestCase, layout=app_layout):
    """Generate module name at run time

    ```python
    Location('...').load(modname=random_name)
    ```

    >>> from importloc import random_name
    >>> Location('app/config.py:Config').load(modname=random_name)
    <class 'u....Config'>
    """


# module name conflict resolution


class R1(TestCase, layout=app_layout):
    """Module name conflict raises error by default

    ```python
    Location('...').load()
    ```

    >>> Location('app/config.py:Config').load()
    <class 'config.Config'>
    >>> Location('app/config.py:Config').load()
    Traceback (most recent call last):
        ...
    importloc.exc.ModuleNameConflict: Module "config" is already imported
    """


class R2(TestCase, layout=app_layout):
    """Reuse module that is already imported

    ```python
    Location('...').load(on_conflict='reuse')
    ```

    >>> C = Location('app/config.py:Config').load()
    >>> C
    <class 'config.Config'>
    >>> old_id = id(C)
    >>> C = Location('app/config.py:Config').load(on_conflict='reuse')
    >>> C
    <class 'config.Config'>
    >>> # C is the same object:
    >>> id(C) == old_id
    True
    """


class R3(TestCase, layout=app_layout):
    """Reload module that is already imported

    ```python
    Location('...').load(on_conflict='reload')
    ```

    >>> import sys
    >>> C = Location('app/config.py:Config').load()
    >>> C
    <class 'config.Config'>
    >>> old_id = id(C)
    >>> mod_id = id(sys.modules['config'])
    >>> C = Location('app/config.py:Config').load(on_conflict='reload')
    >>> C
    <class 'config.Config'>
    >>> # module object remains the same after reloading:
    >>> id(sys.modules['config']) == mod_id
    True
    >>> # C is the new object from reloaded module:
    >>> id(C) == old_id
    False
    """


class R4(TestCase, layout=app_layout):
    """Replace old module with imported one

    ```python
    Location('...').load(on_conflict='replace')
    ```

    >>> import sys
    >>> C = Location('app/config.py:Config').load()
    >>> C
    <class 'config.Config'>
    >>> mod_id = id(sys.modules['config'])
    >>> C = Location('app/config.py:Config').load(on_conflict='replace')
    >>> C
    <class 'config.Config'>
    >>> # module object is the new one:
    >>> id(sys.modules['config']) == mod_id
    False
    """


class R5(TestCase, layout=app_layout):
    """Load module under different generated name

    ```python
    Location('...').load(on_conflict='rename', rename=random_name)
    ```

    >>> from importloc import random_name
    >>> Location('app/config.py').load()
    <module 'config' from ...>
    >>> Location('app/config.py').load(on_conflict='rename', rename=random_name)
    <module 'u...'>
    """


class R6(TestCase, layout=app_layout):
    """Combine override and rename

    ```python
    Location('...').load(modname='...', on_conflict='rename', rename=random_name)
    ```

    >>> from importloc import random_name
    >>> Location('app/config.py').load(modname='app_config')
    <module 'app_config' from ...>
    >>> Location('app/config.py').load(
    ...     modname='app_config', on_conflict='rename', rename=random_name
    ... )
    <module 'u...' from ...>
    """


class O1(TestCase, layout=app_layout):
    """Missing object causes `AttributeError`

    When module was imported but requested object does not exist, `AttributeError`
    is raised.

    >>> Location('app/config.py:unknown').load()
    Traceback (most recent call last):
        ...
    AttributeError: object has no attribute 'unknown'
    >>> # due to import atomicity, module 'config' was removed
    >>> import sys
    >>> 'config' in sys.modules
    False
    """



import re
from typing import Union
from unittest import TestCase

from caseutil import to_kebab
from docsub import click
from importloc import Location, get_subclasses, random_name


@click.group()
def x() -> None:
    pass


@x.command()
@click.argument('subcommand')
@click.argument('testpath', type=click.Path(exists=True, dir_okay=False))
@click.argument('nameregex')
def usage(subcommand: str, testpath: click.Path, nameregex: str) -> None:
    """
    Print usage section based on test cases docstrings
    """
    formatter = {'section': format_section, 'toc': format_toc}[subcommand]
    cases = get_subclasses(Location(str(testpath)).load(random_name), TestCase)
    cases.sort(key=lambda c: c.__firstlineno__)  # type: ignore[attr-defined]
    for case in cases:
        if re.fullmatch(nameregex, case.__name__):
            print(formatter(case.__doc__))


RX_DOCSTRING = re.compile(
    r'^(?P<title>.+?)\n\n(?P<body>.*)',
    re.DOTALL,
)
RX_DOCTEST = re.compile(
    r'(?<=\n)(?P<code>>>>.+?)(?=\n\n|\n$)',
    re.DOTALL,
)


def format_section(text: Union[str, None]) -> str:
    if text is None or (m := RX_DOCSTRING.fullmatch(text)) is None:
        raise ValueError('Invalid docstring format')

    body = m.group('body')
    body = re.sub(r'\n>>>', '\n_Example_\n>>>', body, count=1, flags=re.DOTALL)
    body = RX_DOCTEST.sub(r'```pycon\n\g<code>\n```', body)
    lines = (
        f'### {m.group("title")}\n\n',
        body,
    )
    return ''.join(lines)


def format_toc(text: Union[str, None]) -> str:
    if text is None or (m := RX_DOCSTRING.fullmatch(text)) is None:
        raise ValueError('Invalid docstring format')
    title = m.group('title')
    return f'* [{title}](#{to_kebab(title)})'



from base64 import b32encode
import inspect
from typing import Any, TypeVar, Union
from uuid import uuid4


#: Arbitrary type.
T = TypeVar('T', bound=type)


def get_instances(obj: object, cls: type[T]) -> list[T]:
    """
    Get object member instances of specified type.

    Uses `inspect.getmembers` and `isinstance`.

    >>> import app.plugins
    >>> plugins = get_instances(app.plugins, Plugin)

    :param obj:
        object to get members from.

    :param cls:
        type of members to be returned.

    :return: list of object members.
    """
    return [mem for name, mem in inspect.getmembers(obj) if isinstance(mem, cls)]


def get_subclasses(obj: object, cls: type[T]) -> list[type[T]]:
    """
    Get object member subclasses of specified class, excluding the class itself.

    Uses `inspect.getmembers` and `issubclass`.

    >>> import app.enums
    >>> enums = get_subclasses(app.enums, Enum)

    :param obj:
        object to get members from.

    :param cls:
        type of members to be returned.

    :return: list of object member classes.
    """
    return [
        m
        for name, m in inspect.getmembers(obj)
        if isinstance(m, type) and issubclass(m, cls) and m is not cls
    ]


def getattr_nested(
    obj: object,
    name: str,
    default: Union[type[Exception], Any] = AttributeError,
) -> object:
    """
    Get nested attribute value.

    If attribute chain does not exist, raise exception or return default value.

    >>> from app import config
    >>> options = getattr_nested(config, f'{config.primary}.options')
    >>> missing = getattr_nested(config, 'does.not.exist', None)

    :param obj:
        object to get attribute from.

    :param name:
        dot-separated nested attribute name.

    :param default:
        raise exception if ``default`` is an `Exception` type (`AttributeError` by
        default); otherwise return ``default`` value.

    :raises AttributeError:
        when nested attribute chain does not exist and ``default`` was not adjusted.
    :raises Exception:
        when custom `Exception` passed as ``default``.

    :return:
        nested attribute value or ``default`` value, if specified and not an
        `Exception` subclass.
    """
    current = obj
    for part in name.split('.'):
        try:
            current = getattr(current, part)
        except AttributeError as exc:
            if isinstance(default, type) and issubclass(default, Exception):
                raise default(f'object has no attribute {name!r}') from exc
            else:
                return default
    return current


def random_name(*args: Any, **kwargs: Any) -> str:
    """
    Generate random module name based on UUID4.

    All arguments passed to this function will be ignored.

    >>> random_name()
    'ufoh3xjrrozfcvfheyktg62pzia'
    """
    rand = b32encode(uuid4().bytes).decode('ascii').replace('=', '').lower()
    return f'u{rand}'



r"""
To use any of supported concrete location types, use *generic* `Location` class.
Upon construction, it will return one of *specific* location objects supported:
`ModuleLocation` or `PathLocation`. Alternatively, construct *specific* location types
directly to enforce corresponding location type.

.. list-table::
    :header-rows: 1
    :align: left

    * - Location
      - Format
    * - `PathLocation`
      - ``(?P<path>.*/[^/]*\.py)(:(?P<obj>[^./:]+(?:\.[^./:]+)*))?``
    * - `ModuleLocation`
      - ``(?P<module>[^./:]+(?:\.[^./:]+)*)(:(?P<obj>[^./:]+(?:\.[^./:]+)*))?``

.. list-table::
    :header-rows: 1
    :align: left

    * - Location
      - Examples
    * - `PathLocation`
      - ``svc1/main.py:app``, ``svc1/exceptions.py``, ``../config.py``
    * - `ModuleLocation`
      - ``app.__main__:cli``, ``logging:StreamHandler``

.. raw:: html
    :file: ../../docs/_static/classes-dark.svg
    :class: only-dark

.. raw:: html
    :file: ../../docs/_static/classes-default.svg
    :class: only-light
"""

from abc import ABC
from contextlib import contextmanager
from enum import Enum
import importlib.util
from importlib.machinery import ModuleSpec
from pathlib import Path
import re
import sys
from types import ModuleType
from typing import Any, Callable, Iterable, Literal, Optional, Tuple, TypeVar, Union

from typing_extensions import Self, override

from .exc import InvalidLocation, ModuleNameConflict
from .util import getattr_nested


_OBJ = r'[^./:]+(?:\.[^./:]+)*'
_PYPATH = r'.*/[^/]*\.py'


class ConflictResolution(str, Enum):
    """
    Conflict resolution strategy when module with given name is already imported.
    """

    #: Don't import again, use existing module from `sys.modules`.
    REUSE = 'reuse'
    #: Don't import again, apply `importlib.reload` to existing module in `sys.modules`.
    RELOAD = 'reload'
    #: Delete existing module and use the imported one.
    REPLACE = 'replace'
    #: Retry module import with new generated name, and raise
    #: `~importloc.exc.ModuleNameConflict` exception if conflict appears again.
    RENAME = 'rename'
    #: Raise `~importloc.exc.ModuleNameConflict` exception.
    RAISE = 'raise'


class Location(ABC):
    spec: str
    obj: Optional[str]

    def __new__(cls, spec: Union[str, Path]) -> Union['ModuleLocation', 'PathLocation']:  # type: ignore[misc]
        """
        __init__(self, spec: str) -> Union[ModuleLocation, PathLocation]

        Arbitrary importable location.

        :param spec:
            location specification string.

        :raises InvalidLocation:
            when location string format is incorrect.
        """
        for loctype in cls._types(spec):
            if isinstance(spec, Path) and not spec.is_absolute():
                spec = f'./{spec}'
            match = loctype.match(str(spec))
            if match:
                return loctype(**match.groupdict())
        raise InvalidLocation(spec)

    @classmethod
    def match(cls, spec: str) -> Optional[re.Match[str]]:
        """
        Match location specification string with regular expression for specific
        location type.

        :param spec:
            location specification string.
        """
        raise NotImplementedError

    def load(
        self,
        modname: Union[str, Callable[[Self], str], None] = None,
        on_conflict: Union[ConflictResolution, str] = 'raise',
        rename: Optional[Callable[[str, Self], str]] = None,
    ) -> Union[object, ModuleType]:
        """
        Import requested object or the whole module object from location.

        This operation is atomic:

        * on import error, previous module with the same name is restored
        * on import error, new partially initialized module is removed from `sys.modules`

        :param modname:
            name under which the module will be imported; if `str`,
            use ``modname`` itself; if `~typing.Callable`, use result of
            calling ``modname()`` with current `Location` object;
            otherwise, use default value (see `ModuleLocation.load`
            and `PathLocation.load` for details).

        :param on_conflict:
            behaviour if ``modname`` is already present in `sys.modules`
            (see `ConflictResolution` for details).

        :param rename:
            callable used to generate new module name on name conflict and if
            ``on_conflict`` is ``rename``; first string argument is ``modname``
            that leads to conflict, second argument is current `Location`.

        :raises TypeError | ValueError:
            when passed arguments of wrong type.
        :raises ModuleNameConflict:
            see `ConflictResolution` for details.
        :raises Exception:
            see specific location classes.

        :return:
            `object` when ``obj`` part was specified, otherwise `~types.ModuleType`.
        """
        raise NotImplementedError

    # internal helpers

    @staticmethod
    def _types(spec: Any) -> list[Union[type['ModuleLocation'], type['PathLocation']]]:
        if isinstance(spec, str):
            return [ModuleLocation, PathLocation]
        elif isinstance(spec, Path):
            return [PathLocation]
        else:
            raise TypeError(f'Unexpected spec type {type(spec)}')

    # error helpers with unified error messages

    @staticmethod
    def _args_denied_with_spec() -> ValueError:
        return ValueError('Other arguments are not allowed when spec is passed')

    @staticmethod
    def _arg_required_with_no_spec(arg: str) -> ValueError:
        return ValueError(f'Argument {arg} is required when spec is not passed')

    @staticmethod
    def _import_error(modname: str) -> ImportError:
        return ImportError(f'Module "{modname}" cannot be imported')


class ModuleLocation(Location):
    """
    __init__(self, spec: str) -> None
    __init__(self, *, module: str, obj: Optional[str] = None) -> None

    Package-based importable location, e.g. ``foo.bar:obj``
    """

    module: str
    RX = re.compile(rf'^(?P<module>{_OBJ})(?::(?P<obj>{_OBJ}))?$')

    # bypass Location.__new__
    def __new__(cls, *args: Any, **kwargs: Any) -> 'ModuleLocation':
        return object.__new__(cls)

    def __init__(
        self,
        spec: Optional[str] = None,
        *,
        module: Optional[str] = None,
        obj: Optional[str] = None,
    ) -> None:
        """
        :param spec:
            location specification string; if ``spec`` is passed, other arguments
            must be absent or `None`.

        :param module:
            importable module name; required, if ``spec`` is not passed.

        :param obj:
            dot-separated object name to be imported; when missing, the whole module
            will be loaded

        :raises ValueError:
            when passed incorrect arguments.
        :raises InvalidLocation:
            when location string format is incorrect.
        """
        if spec is None:
            if module is None:
                raise self._arg_required_with_no_spec('module')
            self.module = module
            self.obj = obj
            self.spec = module if self.obj is None else f'{self.module}:{self.obj}'
        else:
            if module is not None or obj is not None:
                raise self._args_denied_with_spec()
            match = self.match(spec)
            if match is None:
                raise InvalidLocation(spec)
            self.spec = spec
            self.module = match.group('module')
            self.obj = match.groupdict().get('obj', None)

    @classmethod
    def match(cls, spec: str) -> Optional[re.Match[str]]:
        """
        Match location specification string with corresponding regular expression.

        :param spec:
            location specification string.
        """
        return cls.RX.match(spec)

    @override
    def load(
        self,
        modname: Union[str, Callable[[Self], str], None] = None,
        on_conflict: Union[ConflictResolution, str] = 'raise',
        rename: Optional[Callable[[str, Self], str]] = None,
    ) -> Union[object, ModuleType]:
        """
        Import requested object or the whole module object from importable module.

        This operation is atomic:

        * on import error, previous module with the same name is restored
        * on import error, new partially initialized module is removed from `sys.modules`

        :param modname:
            name under which the module will be imported; if `str`,
            use ``modname`` itself; if `~typing.Callable`, use result of
            calling ``modname()`` with current `Location` object;
            by default, use ``module`` from ``spec``.

        :param on_conflict:
            behaviour if ``modname`` is already present in `sys.modules`
            (see `ConflictResolution` for details).

        :param rename:
            callable used to generate new module name on name conflict and if
            ``on_conflict`` is ``rename``; first string argument is ``modname`` that
            leads to conflict, second argument is current `Location`.

        :raises TypeError | ValueError:
            when passed arguments of wrong type.
        :raises ModuleNameConflict:
            see `ConflictResolution` for details.
        :raises ModuleNotFoundError:
            when ``modname`` is not discoverable.
        :raises ImportError:
            when module import fails.
        :raises AttributeError:
            when ``obj`` name can't be found in imported module.

        :return:
            `object` when ``obj`` part was specified, otherwise `~types.ModuleType`.
        """
        modname, action = resolve_module_name(
            default=self.module,
            override=modname,
            on_conflict=on_conflict,
            rename=rename,
            loc=self,
        )
        # process
        with atomic_import(modname):
            # import module
            if action == 'import':
                try:
                    modobj = importlib.import_module(modname)
                except ModuleNotFoundError as exc:
                    raise exc
                except Exception as exc:
                    raise self._import_error(modname) from exc
            elif action == 'use':
                modobj = sys.modules[modname]
            else:
                raise RuntimeError('unreachable')
            # get object
            if self.obj is None:
                return modobj
            else:
                return getattr_nested(modobj, self.obj)

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        if self.obj is None:
            return '<{} {!r}>'.format(cls, self.module)
        else:
            return '<{} {!r} obj={!r}>'.format(cls, self.module, self.obj)


class PathLocation(Location):
    """
    __init__(self, spec: str) -> None
    __init__(self, *, path: Union[~pathlib.Path, str], obj: Optional[str] = None) -> None

    Filesystem-based importable location, e.g. ``foo/bar.py:obj``
    """

    path: Path
    RX = re.compile(rf'^(?P<path>{_PYPATH})(?::(?P<obj>{_OBJ}))?$')

    # bypass Location.__new__
    def __new__(cls, *args: Any, **kwargs: Any) -> 'PathLocation':
        return object.__new__(cls)

    def __init__(
        self,
        spec: Union[str, Path, None] = None,
        *,
        path: Union[Path, str, None] = None,
        obj: Optional[str] = None,
    ) -> None:
        """
        :param spec:
            location specification string or `~pathlib.Path` object; if ``spec``
            is passed, other arguments must be absent or `None`.

        :param path:
            path to python source file to import from; required, if ``spec`` is
            not passed.

        :param obj:
            dot-separated object name to be imported; when missing, the whole file
            will be imported as module

        :raises ValueError:
            when passed incorrect arguments.
        :raises InvalidLocation:
            when location string format is incorrect.
        """
        if spec is None:
            if path is None:
                raise self._arg_required_with_no_spec('path')
            self.path = Path(path)
            self.obj = obj
            self.spec = str(path) if self.obj is None else f'{self.path}:{self.obj}'
        else:
            if path is not None or obj is not None:
                raise self._args_denied_with_spec()
            if isinstance(spec, Path):
                self.spec = str(spec)
                self.path = spec
                self.obj = None
            elif isinstance(spec, str):
                match = self.match(spec)
                if match is None:
                    raise InvalidLocation(spec)
                self.spec = spec
                self.path = Path(match.group('path'))
                self.obj = match.groupdict().get('obj', None)
            else:
                raise TypeError(f'Unexpected spec type {type(spec)}')

    @classmethod
    def match(cls, spec: str) -> Optional[re.Match[str]]:
        """
        Match location specification string with corresponding regular expression.

        :param spec:
            location specification string.
        """
        return cls.RX.match(spec)

    @override
    def load(
        self,
        modname: Union[str, Callable[[Self], str], None] = None,
        on_conflict: Union[ConflictResolution, str] = 'raise',
        rename: Optional[Callable[[str, Self], str]] = None,
    ) -> Union[object, ModuleType]:
        """
        Import requested object or the whole module object from location.

        This operation is atomic:

        * on import error, previous module with the same name is restored
        * on import error, new partially initialized module is removed from `sys.modules`

        :param modname:
            name under which the module will be imported; if `str`,
            use ``modname`` itself; if `~typing.Callable`, use result of
            calling ``modname()`` with current `Location` object;
            by default, use ``path`` stem from ``spec``.

        :param on_conflict:
            behaviour if ``modname`` is already present in `sys.modules`
            (see `ConflictResolution` for details).

        :param rename:
            callable used to generate new module name on name conflict and if
            ``on_conflict`` is ``rename``; first string argument is ``modname`` that
            leads to conflict, second argument is current `Location`.

        :raises TypeError | ValueError:
            when passed arguments of wrong type.
        :raises ModuleNameConflict:
            see `ConflictResolution` for details.
        :raises FileNotFoundError:
            when ``path`` does not exist.
        :raises IsADirectoryError:
            when ``path`` is a directory.
        :raises ImportError:
            when module import fails.
        :raises AttributeError:
            when ``obj`` name can't be found in imported module.

        :return:
            `object` when ``obj`` part was specified, otherwise `~types.ModuleType`.
        """
        modname, action = resolve_module_name(
            default=self.path.stem,
            override=modname,
            on_conflict=on_conflict,
            rename=rename,
            loc=self,
        )
        # validate path
        path = self.path.resolve()
        if not path.exists():
            raise FileNotFoundError(f'Path "{path}" does not exist.')
        elif path.is_dir():
            raise IsADirectoryError(f'Path "{path}" is a directory.')
        # load
        with atomic_import(modname):
            # import module
            if action == 'import':
                spec = importlib.util.spec_from_file_location(modname, path)
                if spec is None or spec.loader is None:
                    raise self._import_error(modname)
                try:
                    modobj = load_from_spec(spec)
                except Exception as exc:
                    raise self._import_error(modname) from exc
            elif action == 'use':
                modobj = sys.modules[modname]
            else:
                raise RuntimeError('unreachable')
            # get object
            if self.obj is None:
                return modobj
            else:
                return getattr_nested(modobj, self.obj)

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        if self.obj is None:
            return '<{} {!r}>'.format(cls, str(self.path))
        else:
            return '<{} {!r} obj={!r}>'.format(cls, str(self.path), self.obj)


# undocumented helpers


L = TypeVar('L', bound=Location)


@contextmanager
def atomic_import(modname: str) -> Any:
    old = {m: sys.modules.get(m, None) for m in explode_module_name(modname)}
    try:
        yield
    except:
        for name, value in old.items():
            if value is not None:
                sys.modules[name] = value
            elif name in sys.modules:
                del sys.modules[name]
        raise


def load_from_spec(spec: ModuleSpec) -> ModuleType:
    modobj = importlib.util.module_from_spec(spec)
    modobj.__importloc_spec__ = spec  # type: ignore[attr-defined]
    sys.modules[spec.name] = modobj
    if spec.loader is None:
        raise ImportError(f'Loader not provided for module {spec.name}')
    spec.loader.exec_module(modobj)
    return modobj


def reload(modobj: ModuleType) -> None:
    spec = getattr(modobj, '__importloc_spec__', None)
    if spec:
        spec.loader.exec_module(modobj)
    else:
        importlib.reload(modobj)


def explode_module_name(modname: str) -> Iterable[str]:
    end = 0
    while end != -1:
        end = modname.find('.', end + 1)
        yield modname if end == -1 else modname[:end]


def resolve_module_name(
    default: str,
    override: Union[str, Callable[[L], str], None],
    on_conflict: Union[ConflictResolution, str],
    rename: Optional[Callable[[str, L], str]],
    loc: L,
) -> Tuple[str, Literal['use', 'import']]:
    # validate args
    on_conflict = ConflictResolution(on_conflict)
    if on_conflict == ConflictResolution.RENAME and not callable(rename):
        raise ValueError('rename must be callable')

    # determine initial module name
    if override is None:
        modname = default
    elif isinstance(override, str):
        modname = override
    elif callable(override):
        modname = override(loc)
    else:
        raise ValueError(f'Unexpected modname override type {type(override)}')

    # module already imported?
    if modname not in sys.modules:
        return modname, 'import'

    # resolve
    if on_conflict == ConflictResolution.REUSE:
        return modname, 'use'
    elif on_conflict == ConflictResolution.RELOAD:
        reload(sys.modules[modname])
        return modname, 'use'
    elif on_conflict == ConflictResolution.REPLACE:
        return modname, 'import'
    elif on_conflict == ConflictResolution.RENAME:
        modname = rename(modname, loc)  # type: ignore # checked above
        if modname in sys.modules:
            raise ModuleNameConflict(modname)
        return modname, 'import'
    elif on_conflict == ConflictResolution.RAISE:
        raise ModuleNameConflict(modname)
    else:
        raise RuntimeError('unreachable')
